class InvalidComponentError(Exception):
    """Exception raised for invalid components in the system."""

    def __init__(
        self,
        component_name: str = "Unknown Component",
        message: str = "Component invalid - no reason given",
    ):
        self.component_name = component_name
        self.message = f"{component_name}: {message}"
        super().__init__(self.message)


class ToolCallError(Exception):
    """Exception raised for errors during tool calls."""

    def __init__(
        self,
        tool_name: str = "Unknown Tool",
        message: str = "Tool call failed - no reason given",
    ):
        self.tool_name = tool_name
        self.message = f"{self.tool_name}: {message}"
        super().__init__(self.message)



import asyncio
import logging
import signal
from typing import Callable

from daystrom.components.base import Agent
from daystrom.components.openai import OpenAIChatCompletions
from daystrom.providers import Provider

from .transports.base import Transport

log = logging.getLogger(__name__)


class Computer:
    def __init__(self, agent_factory: Callable[[], Agent]):
        self.agent_factory = agent_factory
        self.transports: list[Transport] = []
        self.sessions: dict[str, Agent] = {}
        self._stop_event = asyncio.Event()

    def register(self, transport: Transport):
        transport.computer = self
        self.transports.append(transport)

    def get_or_create_agent(self, session_id: str) -> Agent:
        if session_id not in self.sessions:
            self.sessions[session_id] = self.agent_factory()
        return self.sessions[session_id]

    async def handle_message(self, session_id: str, text: str) -> str:
        log.info(f"Received message for session {session_id}: {text[:50] + "..." if len(text) > 50 else ""}")

        parts = text.strip().split()
        command = parts[0] if parts else ""
        args = parts[1:]
        match command:
            case "/new-session":
                model_dict = {
                    "claude": "anthropic/claude-opus-4.6",
                    "gpt": "openai/gpt-5.2",
                    "minimax": "minimax/minimax-m2.5",
                    "kimi": "moonshotai/kimi-k2.5",
                    "glm": "z-ai/glm-5",
                }
                if args:
                    model = model_dict.get(args[0])
                    if model:
                        def agent_factory():
                            llm = OpenAIChatCompletions(provider=Provider.OPENROUTER, model=model)
                            agent = Agent(llm=llm)
                            return agent
                        self.agent_factory = agent_factory
                        log.info(f"Agent factory reset with model: {model}")

                self.sessions.pop(session_id, None)
                return "Session reset."

        agent = self.get_or_create_agent(session_id)
        response = await asyncio.to_thread(agent.invoke, text)
        return response.text

    def run(self):
        asyncio.run(self._run())

    async def _run(self):
        loop = asyncio.get_running_loop()
        for sig in (signal.SIGINT, signal.SIGTERM):
            loop.add_signal_handler(sig, self._stop_event.set)

        started_transports: list[Transport] = []
        try:
            for transport in self.transports:
                await transport.start()
                started_transports.append(transport)

            log.info("Computer is running")
            await self._stop_event.wait()
            log.info("Shutting down")
        finally:
            # Ensure all started transports are stopped, even if startup or run fails.
            for transport in reversed(started_transports):
                try:
                    await transport.stop()
                except Exception:
                    log.exception("Error while stopping transport %r", transport)



import logging

try:
    from telegram import Update
    from telegram.ext import Application, MessageHandler, filters
except ImportError:
    raise ImportError(
        "python-telegram-bot is required for TelegramTransport. "
        "Install with: uv add daystrom[telegram] (or run with --extra telegram)"
    )

from .base import Transport

log = logging.getLogger(__name__)

TELEGRAM_MESSAGE_LIMIT = 4096


class TelegramTransport(Transport):
    def __init__(self, token: str):
        self.token = token
        self._app: Application | None = None

    async def start(self):
        self._app = Application.builder().token(self.token).build()

        if not self._app or not self._app.updater:
            raise RuntimeError("Failed to create Telegram application")

        self._app.add_handler(MessageHandler(filters.TEXT, self._on_message))
        await self._app.initialize()
        await self._app.start()
        await self._app.updater.start_polling()
        log.info("Telegram transport started")

    async def stop(self):
        if self._app and self._app.updater:
            await self._app.updater.stop()
            await self._app.stop()
            await self._app.shutdown()
            log.info("Telegram transport stopped")

    async def _on_message(self, update: Update, _context):
        if not update.message or not update.message.text or not update.effective_chat:
            return

        session_id = f"telegram:{update.effective_chat.id}"
        response = await self.computer.handle_message(session_id, update.message.text)

        if not response:
            return

        for chunk in _split_message(response):
            await update.message.reply_text(chunk)


def _split_message(text: str) -> list[str]:
    if len(text) <= TELEGRAM_MESSAGE_LIMIT:
        return [text]

    chunks = []
    while text:
        if len(text) <= TELEGRAM_MESSAGE_LIMIT:
            chunks.append(text)
            break

        split_at = text.rfind("\n", 0, TELEGRAM_MESSAGE_LIMIT)
        if split_at == -1:
            split_at = TELEGRAM_MESSAGE_LIMIT

        chunks.append(text[:split_at])
        text = text[split_at:].lstrip("\n")

    if not chunks:
        chunks.append("Something went wrong, no response from the Computer")

    return chunks



import json
import os
from typing import get_origin

from openai import OpenAI
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionDeveloperMessageParam,
    ChatCompletionMessageFunctionToolCall,
    ChatCompletionMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionToolParam,
    ChatCompletionToolUnionParam,
    ChatCompletionUserMessageParam,
)
from openai.types.shared_params import FunctionDefinition

from daystrom import Provider
from daystrom.components import LLM, Context, LLMResponse, Tool, ToolCall
from daystrom.exceptions import InvalidComponentError


class OpenAIChatCompletions(LLM):
    def __init__(
        self,
        model: str,
        provider: Provider | None = None,
        api_key: str | None = None,
        tools: dict[str, Tool] | None = None,
    ):
        super().__init__(
            model=model,
            tools=tools or {},
            provider=provider or Provider.OPENAI,
        )
        self.client = OpenAI(
            base_url=self.provider.value.base_url,
            api_key=api_key or self.provider.value.get_api_key(),
        )

    def invoke(self, context: Context | None = None) -> LLMResponse:
        messages = self._get_prompt_context(context)
        completion = self.client.chat.completions.create(
            model=self.model,
            tools=self._get_tool_context(),
            messages=messages,
        )
        self.track_usage(completion.usage)
        completion_text = completion.choices[0].message.content or ""

        # tool calls from openai api
        tool_calls = []
        for tool_call in completion.choices[0].message.tool_calls or []:
            if isinstance(tool_call, ChatCompletionMessageFunctionToolCall):
                tool = ToolCall(
                    tool=self.tools[tool_call.function.name],
                    tool_call_id=tool_call.id,
                    args=[],
                    kwargs=(
                        json.loads(tool_call.function.arguments)
                        if tool_call.function.arguments
                        else {}
                    ),
                )
                tool_calls.append(tool)
            else:
                raise InvalidComponentError(
                    self.__class__.__name__,
                    "Found unsupported tool call - missing 'function' attribute",
                )

        response = LLMResponse(text=completion_text, tool_calls=tool_calls)
        return response

    def track_usage(self, usage):
        if usage:
            self.output_tokens += usage.completion_tokens
            self.input_tokens += usage.prompt_tokens

    def _get_prompt_context(
        self, context: Context | None = None
    ) -> list[ChatCompletionMessageParam]:
        """
        Returns the messages in the context formatted for OpenAI API
        """
        if not context:
            return []

        fmt_messages = []
        for msg in context.messages:
            match msg.role:
                case "user":
                    fmt_messages.append(
                        ChatCompletionUserMessageParam(role="user", content=msg.text)
                    )
                case "assistant":
                    tool_calls = []
                    for tool_call in msg.tool_calls:
                        # ChatCompletionMessageToolCallUnionParam
                        tool_calls.append(
                            {
                                "function": {
                                    "name": tool_call.tool.name,
                                    "arguments": json.dumps(tool_call.kwargs),
                                },
                                "type": "function",
                                "id": tool_call.tool_call_id,
                            }
                        )
                    if tool_calls:
                        fmt_messages.append(
                            ChatCompletionAssistantMessageParam(
                                role="assistant",
                                content=msg.text,
                                tool_calls=tool_calls,
                            )
                        )
                    else:
                        fmt_messages.append(
                            ChatCompletionAssistantMessageParam(
                                role="assistant", content=msg.text
                            )
                        )
                case "system":
                    fmt_messages.append(
                        ChatCompletionDeveloperMessageParam(
                            role="developer", content=msg.text
                        )
                    )
                case "tool":
                    fmt_messages.append(
                        ChatCompletionToolMessageParam(
                            role="tool", content=msg.text, tool_call_id=msg.tool_call_id
                        )
                    )
                case _:
                    raise ValueError(
                        f"Unsupported message role: {msg.role} for {self.__class__.__name__}"
                    )

        return fmt_messages

    def _get_tool_context(self) -> list[ChatCompletionToolUnionParam]:
        tool_schemas = []

        for tool in self.tools.values():
            function = FunctionDefinition(
                name=tool.name,
                description=tool.description,
                parameters=self._format_tool_params(tool),
            )
            tool_schema = ChatCompletionToolParam(function=function, type="function")
            tool_schemas.append(tool_schema)
        return tool_schemas

    def _format_tool_params(self, tool: Tool) -> dict:
        params = {"type": "object", "properties": {}}
        required_params = []
        type_map = {
            dict: "object",
            list: "array",
            tuple: "array",
            str: "string",
            int: "integer",
            float: "number",
            None: "null",
            bool: "boolean",
        }

        for pname, pinfo in tool.params.items():
            params["properties"][pname] = {
                "type": type_map[get_origin(pinfo["type"]) or pinfo["type"]],
                "description": pinfo["description"],
            }

            param_items = pinfo.get("items")
            if param_items is not None:
                param_type = param_items["type"]
                params["properties"][pname]["items"] = {
                    "type": type_map[get_origin(param_type) or param_type],
                }

            if pinfo["required"]:
                required_params.append(pname)

        if required_params:
            params["required"] = required_params

        return params



import functools
import inspect
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, TypeVar, get_origin

from docstring_parser import parse

from daystrom import Provider
from daystrom.exceptions import ToolCallError

ComponentResponseT = TypeVar("ComponentResponseT")

log = logging.getLogger(__name__)


class Component(Generic[ComponentResponseT], ABC):
    @abstractmethod
    def invoke(self, *args, **kwargs) -> ComponentResponseT | None:
        pass  # pragma: no cover


class Tool:
    def __init__(
        self,
        callable,
        name: str = "",
        display_name: str = "",
        description: str = "",
        params: dict | None = None,
    ):
        self.callable = callable
        self.name = name or callable.__name__
        self.display_name = display_name or self.name.replace("_", " ").title()
        self.description = description or callable.__doc__ or ""
        self.params = params or {}

    def __str__(self):
        return f"Tool(name={self.name}, description={self.description}, params={self.params})"

    def call(self, *args, **kwargs):
        return self.callable(*args, **kwargs)


@dataclass
class ToolCall:
    tool: Tool
    tool_call_id: str
    args: list
    kwargs: dict


class Message:
    def __init__(
        self,
        role: str,
        text: str,
        tool_call_id: str = "",
        tool_calls: list[ToolCall] | None = None,
    ):
        self.role = role
        self.text = text
        self.tool_call_id = tool_call_id
        self.tool_calls: list[ToolCall] = tool_calls or []

    def __str__(self):
        parts = []
        parts.append(f"{self.role}: {self.text}")
        if self.tool_call_id:
            parts.append(f"    Tool Call ID: {self.tool_call_id}")
        if self.tool_calls:
            parts.append("    Tool Calls:")
            for tool in self.tool_calls:
                parts.append(f"    {str(tool)}")

        return "\n".join(parts)


class Context:
    def __init__(self, messages: list | None = None):
        self.messages: list[Message] = messages or []

    def add_message(
        self,
        role: str,
        text: str,
        tool_call_id: str = "",
        tool_calls: list[ToolCall] | None = None,
    ):
        self.messages.append(
            Message(
                text=text, role=role, tool_call_id=tool_call_id, tool_calls=tool_calls
            )
        )

    def print_feed(self):
        for message in self.messages:
            print(message)


@dataclass
class LLMResponse:
    text: str
    tool_calls: list[ToolCall]


@dataclass
class AgentResponse:
    text: str


class LLM(Component[LLMResponse]):
    tools: dict[str, Tool]
    provider: Provider
    model: str
    input_tokens: int
    output_tokens: int
    input_cost: float | None
    output_cost: float | None
    context_limit: int | None
    output_limit: int | None

    def __init__(
        self,
        provider: Provider,
        model: str,
        tools: dict[str, Tool] | None = None,
    ):
        self.provider = provider
        self.model = model

        self.tools = tools or {}
        self.input_tokens: int = 0
        self.output_tokens: int = 0

        self.input_cost = None
        self.output_cost = None
        self.context_limit = None
        self.output_limit = None
        model_metadata = provider.value.models.get(self.model)
        if model_metadata:
            self.input_cost = model_metadata.input_cost
            self.output_cost = model_metadata.output_cost
            self.context_limit = model_metadata.context_limit
            self.output_limit = model_metadata.output_limit

    @abstractmethod
    def invoke(self, *args, **kwargs) -> LLMResponse:
        pass  # pragma: no cover

    @abstractmethod
    def track_usage(self, *args, **kwargs):
        pass  # pragma: no cover

    @property
    def total_cost(self) -> float:
        """Calculate total cost based on token usage and costs.

        Returns 0.0 if cost information is not available.
        """
        if self.input_cost is None and self.output_cost is None:
            return 0.0
        # Costs are per million tokens
        input_total = (self.input_tokens / 1_000_000) * (self.input_cost or 0.0)
        output_total = (self.output_tokens / 1_000_000) * (self.output_cost or 0.0)
        return input_total + output_total


DEFAULT_TOOLS = {}


# this is a decorator to be @tool above each tool function
def tool(func):
    docstring = parse(func.__doc__ or "")
    inspect_params = inspect.signature(func).parameters

    func_params = {}

    for idx, (name, param) in enumerate(inspect_params.items()):
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            raise TypeError("*args is not supported in tool parameters.")
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            raise TypeError("**kwargs is not supported in tool parameters.")

        required = False
        if param.default is inspect.Parameter.empty:
            required = True

        description = ""
        if len(docstring.params) >= idx + 1:
            description = docstring.params[idx].description

        func_params[name] = {
            "type": param.annotation,
            "description": description,
            "required": required,
        }

        if get_origin(param.annotation) in (list, tuple):
            if len(param.annotation.__args__) > 1:
                raise TypeError(
                    "Only single-type iterables are allowed as parameters to tool calls."
                )
            else:
                func_params[name]["items"] = {"type": param.annotation.__args__[0]}

    tool_desc = docstring.long_description or docstring.short_description or ""

    new_tool = Tool(func, name=func.__name__, description=tool_desc, params=func_params)

    DEFAULT_TOOLS[new_tool.name] = new_tool

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


class Agent(Component[AgentResponse]):
    llm: LLM
    context: Context
    max_loops: int
    tools: dict

    def __init__(
        self,
        llm: LLM,
        context: Context | None = None,
        tools: dict[str, Tool] | None = None,
        max_loops: int = 30,
    ):
        self.llm = llm

        if context:
            self.context = context
        else:
            self.context = Context()

        self.max_loops = max_loops
        self.tools = tools or DEFAULT_TOOLS
        self.llm.tools = self.tools

    def invoke(self, prompt, *args, **kwargs) -> AgentResponse:
        loop = 0
        self.context.add_message("user", prompt)

        res = None
        while loop < self.max_loops:
            loop += 1
            res = self.llm.invoke(self.context)
            self.context.add_message(
                "assistant", text=res.text, tool_calls=res.tool_calls
            )

            # if no tools were called, the agent loop is done
            if not res.tool_calls:
                break

            for tool_call in res.tool_calls:
                try:
                    tool_res = tool_call.tool.call(*tool_call.args, **tool_call.kwargs)
                    self.context.add_message(
                        "tool", text=tool_res, tool_call_id=tool_call.tool_call_id
                    )
                except ToolCallError as e:
                    self.context.add_message(
                        "tool",
                        text=f"Tool call failed! Error: {e.message}",
                        tool_call_id=tool_call.tool_call_id,
                    )
                    log.exception(f"Tool call failed: {tool_call.tool.name}")
                except Exception as e:
                    self.context.add_message(
                        "tool",
                        text=f"Tool call failed! Error: {e}",
                        tool_call_id=tool_call.tool_call_id,
                    )
                    log.exception(f"Tool call failed: {tool_call.tool.name}")

        agent_res = AgentResponse(text=(res.text if res else ""))
        return agent_res



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from daystrom.server.computer import Computer


class Transport(ABC):
    computer: Computer

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass



from .base import (
    DEFAULT_TOOLS,
    LLM,
    Agent,
    AgentResponse,
    Component,
    Context,
    LLMResponse,
    Message,
    Tool,
    ToolCall,
    tool,
)
from .instructor import Instructor

__all__ = [
    "DEFAULT_TOOLS",
    "LLM",
    "Agent",
    "AgentResponse",
    "Component",
    "Context",
    "LLMResponse",
    "Message",
    "Tool",
    "ToolCall",
    "tool",
    "Instructor",
]



import os

from openrouter import OpenRouter
from openrouter.components import AssistantMessage
from openrouter.components import Message as OpenRouterMessage  # , ToolResponseMessage
from openrouter.components import SystemMessage, UserMessage

from daystrom import Provider
from daystrom.components import LLM, Context, LLMResponse


class OpenRouterChat(LLM):
    def __init__(
        self,
        model: str,
        api_key: str | None = None,
    ):
        super().__init__(model=model, provider=Provider.OPENROUTER)
        self.client = OpenRouter(api_key=api_key or self.provider.value.get_api_key())

    def invoke(self, context: Context | None = None) -> LLMResponse:
        response = LLMResponse(text="".join(self.invoke_stream(context)), tool_calls=[])
        return response

    def invoke_stream(self, context: Context | None = None):
        messages = self._get_prompt_context(context)
        res = self.client.chat.send(messages=messages, model=self.model, stream=True)
        # breakpoint()

        response_content = ""
        for event in res:
            self.track_usage(event.usage)
            if event.choices and isinstance(event.choices[0].delta.content, str):
                content_chunk = event.choices[0].delta.content
                response_content += content_chunk
                yield content_chunk

    async def ainvoke(self, context: Context | None = None) -> LLMResponse:
        response = LLMResponse(
            text="".join([chunk async for chunk in self.ainvoke_stream(context)]),
            tool_calls=[],
        )
        return response

    async def ainvoke_stream(self, context: Context | None = None):
        messages = self._get_prompt_context(context)
        res = await self.client.chat.send_async(
            messages=messages, model=self.model, stream=True
        )

        response_content = ""
        async for event in res:
            self.track_usage(event.usage)
            if event.choices and isinstance(event.choices[0].delta.content, str):
                content_chunk = event.choices[0].delta.content
                response_content += content_chunk
                yield content_chunk

    def track_usage(self, usage):
        if usage:
            self.output_tokens += usage.completion_tokens
            self.input_tokens += usage.prompt_tokens

    def _get_prompt_context(self, context: Context | None) -> list[OpenRouterMessage]:
        """
        Returns the messages in the context formatted for OpenRouter API
        """
        if not context:
            return []

        fmt_messages = []
        for msg in context.messages:
            match msg.role:
                case "user":
                    fmt_messages.append(UserMessage(role=msg.role, content=msg.text))
                case "assistant":
                    fmt_messages.append(
                        AssistantMessage(role=msg.role, content=msg.text)
                    )
                case "system":
                    fmt_messages.append(SystemMessage(role=msg.role, content=msg.text))

        return fmt_messages



import logging
import os
from logging.config import dictConfig

log = logging.getLogger("daystrom")

if not log.hasHandlers():
    dictConfig(
        {
            "version": 1,
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "standard",
                }
            },
            "formatters": {
                "standard": {
                    "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                },
                "verbose": {
                    "format": "%(asctime)s - %(module)s - %(levelname)s - %(message)s (PID: %(process)d) (%(pathname)s:%(lineno)d)"
                },
            },
            "root": {
                "handlers": ["console"],
                "level": "WARNING",
            },
            "loggers": {
                "daystrom": {
                    "handlers": ["console"],
                    "level": os.getenv("DAYSTROM_LOG_LEVEL", "INFO"),
                    "propagate": False,
                },
            },
        }
    )



import logging
import os
from dataclasses import dataclass
from enum import Enum
from functools import cache

import httpx

log = logging.getLogger(__name__)


@dataclass
class ModelMetadata:
    name: str
    display_name: str
    input_cost: float | None
    """cost per million tokens"""
    output_cost: float | None
    """cost per million tokens"""
    context_limit: int | None
    output_limit: int | None


class ProviderMetadata:
    name: str
    display_name: str
    base_url: str | None
    models: dict[str, ModelMetadata]
    env_names: list[str]

    def __init__(
        self,
        name: str,
        display_name: str,
        base_url: str | None = None,
    ) -> None:
        self.name = name
        self.display_name = display_name
        self.base_url = base_url
        self.models: dict[str, ModelMetadata] = {}
        self.env_names = []

        base_data = self.get_data()
        if base_data:
            provider_data = base_data.get(self.name)
            if provider_data:
                self.env_names = provider_data.get("env", [])
                for model, metadata in provider_data["models"].items():
                    cost = metadata.get("cost")
                    limit = metadata.get("limit")
                    model_metadata = ModelMetadata(
                        name=metadata.get("id"),
                        display_name=metadata.get("name"),
                        input_cost=cost.get("input") if cost else None,
                        output_cost=cost.get("output") if cost else None,
                        context_limit=limit.get("context") if limit else None,
                        output_limit=limit.get("output") if limit else None,
                    )
                    self.models[model] = model_metadata

    @classmethod
    @cache
    def get_data(cls) -> dict | None:
        metadata_api_url = "https://models.dev/api.json"
        data = None
        try:
            response = httpx.get(metadata_api_url, timeout=3.0)
            response.raise_for_status()
            data = response.json()
        except Exception:
            err = "Failed to pull model metadata from models.dev. Proceeding without..."
            log.error(err)
        return data

    def get_api_key(self):
        """
        For straightforward providers that just need an API key, this gets the first one found in the list from models.dev

        More complicated providers (e.g. AWS Bedrock) will need to manage this in their own component
        """
        api_key = None
        for env_name in self.env_names:
            api_key = os.getenv(env_name)
            if api_key:
                break
        return api_key


class Provider(Enum):
    OPENAI = ProviderMetadata(
        name="openai",
        display_name="Open AI",
        base_url="https://api.openai.com/v1",
    )
    OPENROUTER = ProviderMetadata(
        name="openrouter",
        display_name="OpenRouter",
        base_url="https://openrouter.ai/api/v1",
    )
    AWS_BEDROCK = ProviderMetadata(
        name="amazon-bedrock",
        display_name="AWS Bedrock",
    )



import json
import os
from typing import TypeVar

import instructor
from openai.types.chat import (  # ChatCompletionDeveloperMessageParam, # should probably use this one, it replaces system_message on some newer models apparently; ChatCompletionFunctionMessageParam,; ChatCompletionToolMessageParam,
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)
from pydantic import BaseModel

from daystrom import Provider
from daystrom.components import Component, Context
from daystrom.exceptions import InvalidComponentError

InstructorResponseT = TypeVar("InstructorResponseT", bound=BaseModel)


class Instructor(Component[InstructorResponseT]):
    client: instructor.Instructor
    response_model: type[InstructorResponseT]
    context: Context

    def __init__(
        self,
        provider: Provider,
        model: str,
        response_model: type[InstructorResponseT],
        api_key: str | None = None,
        context: Context | None = None,
    ):
        self.provider = provider
        self.model = model
        self.client = instructor.from_provider(
            f"{provider.value.name}/{model}",
            api_key=api_key or self.provider.value.get_api_key(),
        )

        if not self.client:
            raise InvalidComponentError(
                self.__class__.__name__, f"Unsupported provider: {provider.name}"
            )

        self.response_model = response_model
        if context:
            self.context = context
        else:
            self.context = Context()

    def invoke(self, prompt: str) -> InstructorResponseT:
        self.context.add_message("user", prompt)
        messages = self._get_prompt_context()
        response = self.client.create(
            response_model=self.response_model, messages=messages, max_retries=3
        )
        return response

    def _get_prompt_context(self) -> list[ChatCompletionMessageParam]:
        """
        Returns the messages in the context formatted for OpenRouter API
        """
        fmt_messages = []
        for msg in self.context.messages:
            match msg.role:
                case "user":
                    fmt_messages.append(
                        ChatCompletionUserMessageParam(role="user", content=msg.text)
                    )
                case "assistant":
                    tool_calls = []
                    for tool_call in msg.tool_calls:
                        # ChatCompletionMessageToolCallUnionParam
                        tool_calls.append(
                            {
                                "function": {
                                    "name": tool_call.tool.name,
                                    "arguments": json.dumps(tool_call.kwargs),
                                },
                                "type": "function",
                                "id": tool_call.tool_call_id,
                            }
                        )
                    if tool_calls:
                        fmt_messages.append(
                            ChatCompletionAssistantMessageParam(
                                role="assistant",
                                content=msg.text,
                                tool_calls=tool_calls,
                            )
                        )
                    else:
                        fmt_messages.append(
                            ChatCompletionAssistantMessageParam(
                                role="assistant", content=msg.text
                            )
                        )
                case "system":
                    fmt_messages.append(
                        ChatCompletionSystemMessageParam(
                            role="system", content=msg.text
                        )
                    )
                case "tool":
                    fmt_messages.append(
                        ChatCompletionToolMessageParam(
                            role="tool", content=msg.text, tool_call_id=msg.tool_call_id
                        )
                    )
                case _:
                    raise ValueError(
                        f"Unsupported message role: {msg.role} for {self.__class__.__name__}"
                    )
        return fmt_messages



from .base import Transport

__all__ = ["Transport"]

try:
    from .telegram import TelegramTransport
except ImportError:
    TelegramTransport = None  # Optional; not available without telegram extra
else:
    __all__.append("TelegramTransport")



import json

from markdownify import markdownify as md

from daystrom.components import tool
from daystrom.exceptions import ToolCallError


@tool
def web_fetch(url: str, format: str = "markdown") -> str:
    """Fetches content from a given URL.

    Args:
        url (str): The URL to fetch content from.
        format (str, optional): The format of the content to fetch, text, html, json, or markdown. Default "markdown".

    Returns:
        str: The fetched content as a string.
    """
    import httpx

    accept_header = "*/*"
    match format:
        case "text":
            accept_header = (
                "text/plain;q=1.0, text/markdown;q=0.9, text/html;q=0.8, */*;q=0.1"
            )
        case "html":
            accept_header = "text/html;q=1.0, application/xhtml+xml;q=0.9, text/plain;q=0.8, text/markdown;q=0.7, */*;q=0.1"
        case "json":
            accept_header = "application/json;q=1.0, text/markdown;q=0.9, text/x-markdown;q=0.8, text/plain;q=0.7, text/html;q=0.6, */*;q=0.1"
        case "markdown":
            accept_header = "text/markdown;q=1.0, text/x-markdown;q=0.9, text/plain;q=0.8, text/html;q=0.7, */*;q=0.1"
        case _:
            raise ToolCallError("web_fetch", f"Unsupported format: {format}")

    headers = {
        "accept": accept_header,
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Safari/537.36",
    }
    response = httpx.get(url, timeout=10.0, headers=headers, follow_redirects=False)
    if response.is_redirect:
        redirect_url = response.headers.get("location")
        raise ToolCallError(
            "web_fetch",
            f"Redirection not supported by tool - to get the content, call the web_fetch tool again with the new URL: {redirect_url}",
        )

    response.raise_for_status()
    response_type = response.headers.get("content-type") or ""

    match format:
        case "text":
            ans = response.text
            if "text/html" in response_type:
                ans = md(ans, convert=[])
            return ans
        case "html":
            return response.text
        case "json":
            ans = response.text
            if "application/json" in response_type:
                ans = json.dumps(response.json())
            return ans
        case "markdown":
            ans = response.text
            if "text/html" in response_type:
                ans = md(ans)
            return ans
        case _:
            raise ToolCallError("web_fetch", f"Unsupported format: {format}")



import os
from dataclasses import asdict, dataclass
from typing import Literal, get_origin

import boto3
from botocore.config import Config

from daystrom import Provider
from daystrom.components import LLM, Context, LLMResponse, Tool, ToolCall
from daystrom.exceptions import InvalidComponentError


@dataclass
class AwsAccessKeyAuth:
    kind: Literal["access_key"] = "access_key"
    config: Config = Config(signature_version="v4")
    aws_access_key_id: str = ""
    aws_secret_access_key: str = ""
    aws_session_token: str | None = None
    region_name: str = ""

    def get_auth_dict(self) -> dict:
        auth_dict = asdict(self)
        auth_dict = {
            k: v for k, v in auth_dict.items() if k != "kind" and v is not None
        }
        # breakpoint()
        return auth_dict


@dataclass
class AwsBearerTokenAuth:
    kind: Literal["bearer_token"] = "bearer_token"
    aws_bearer_token_bedrock: str = ""
    region_name: str = ""

    def get_auth_dict(self) -> dict:
        auth_dict = asdict(self)
        # boto3.client doesn't accept aws_bearer_token_bedrock as a direct
        # arg like it does access and secret keys, so we have to omit it
        # here and let it pull it from the env
        auth_dict = {
            k: v
            for k, v in auth_dict.items()
            if k != "kind" and k != "aws_bearer_token_bedrock" and v is not None
        }
        return auth_dict


@dataclass
class AwsFailedAuth:
    kind: Literal["failed"] = "failed"
    message: str = ""


class BedrockConverse(LLM):
    def __init__(
        self,
        model: str,
        region_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        aws_bearer_token_bedrock: str | None = None,
        tools: dict[str, Tool] | None = None,
    ):
        super().__init__(
            model=model,
            tools=tools or {},
            provider=Provider.AWS_BEDROCK,
        )

        auth_obj = self.get_auth(
            region_name,
            aws_access_key_id,
            aws_secret_access_key,
            aws_session_token,
            aws_bearer_token_bedrock,
        )
        if auth_obj.kind == "failed":
            raise InvalidComponentError(self.__class__.__name__, auth_obj.message)
        self.client = boto3.client(
            "bedrock-runtime",
            **auth_obj.get_auth_dict(),
        )

    def invoke(self, context: Context | None = None) -> LLMResponse:
        system_messages, messages = self._get_prompt_context(context)

        kwargs = {
            "modelId": self.model,
            "messages": messages,
        }
        if system_messages:
            kwargs["system"] = system_messages

        tool_config = self._get_tool_context()
        if tool_config:
            kwargs["toolConfig"] = tool_config

        response = self.client.converse(**kwargs)
        self.track_usage(response.get("usage"))

        # Parse response content blocks
        content_blocks = response["output"]["message"].get("content", [])

        text = ""
        tool_calls = []
        for block in content_blocks:
            if "text" in block:
                text += block["text"]
            elif "toolUse" in block:
                tool_use = block["toolUse"]
                tool_calls.append(
                    ToolCall(
                        tool=self.tools[tool_use["name"]],
                        tool_call_id=tool_use["toolUseId"],
                        args=[],
                        kwargs=tool_use.get("input", {}),
                    )
                )

        return LLMResponse(text=text, tool_calls=tool_calls)

    def track_usage(self, usage):
        if usage:
            self.input_tokens += usage.get("inputTokens", 0)
            self.output_tokens += usage.get("outputTokens", 0)

    def _get_prompt_context(
        self, context: Context | None = None
    ) -> tuple[list[dict], list[dict]]:
        """
        Returns (system_messages, conversation_messages) formatted for the
        Bedrock Converse API.

        System messages are separated out because Converse takes them as a
        top-level parameter rather than inline in the messages array.

        Consecutive tool-result messages are grouped into a single 'user'
        message with multiple toolResult content blocks, since Converse
        requires alternating user/assistant roles.
        """
        if not context:
            return [], []

        system_messages: list[dict] = []
        messages: list[dict] = []

        i = 0
        while i < len(context.messages):
            msg = context.messages[i]

            match msg.role:
                case "system":
                    system_messages.append({"text": msg.text})
                    i += 1
                case "user":
                    messages.append({"role": "user", "content": [{"text": msg.text}]})
                    i += 1
                case "assistant":
                    content: list[dict] = []
                    if msg.text:
                        content.append({"text": msg.text})
                    for tool_call in msg.tool_calls:
                        content.append(
                            {
                                "toolUse": {
                                    "toolUseId": tool_call.tool_call_id,
                                    "name": tool_call.tool.name,
                                    "input": tool_call.kwargs,
                                }
                            }
                        )
                    messages.append({"role": "assistant", "content": content})
                    i += 1
                case "tool":
                    # Group consecutive tool messages into one user message
                    tool_results: list[dict] = []
                    while (
                        i < len(context.messages) and context.messages[i].role == "tool"
                    ):
                        tool_msg = context.messages[i]
                        tool_results.append(
                            {
                                "toolResult": {
                                    "toolUseId": tool_msg.tool_call_id,
                                    "content": [{"text": tool_msg.text}],
                                }
                            }
                        )
                        i += 1
                    messages.append({"role": "user", "content": tool_results})
                case _:
                    raise ValueError(
                        f"Unsupported message role: {msg.role} for {self.__class__.__name__}"
                    )

        return system_messages, messages

    def _get_tool_context(self) -> dict:
        if not self.tools:
            return {}

        tools = []
        for tool in self.tools.values():
            tool_spec = {
                "toolSpec": {
                    "name": tool.name,
                    "description": tool.description or tool.name,
                    "inputSchema": {"json": self._format_tool_params(tool)},
                }
            }
            tools.append(tool_spec)

        return {"tools": tools}

    def _format_tool_params(self, tool: Tool) -> dict:
        params: dict = {"type": "object", "properties": {}}
        required_params = []
        type_map = {
            dict: "object",
            list: "array",
            tuple: "array",
            str: "string",
            int: "integer",
            float: "number",
            None: "null",
            bool: "boolean",
        }

        for pname, pinfo in tool.params.items():
            params["properties"][pname] = {
                "type": type_map[get_origin(pinfo["type"]) or pinfo["type"]],
                "description": pinfo["description"],
            }

            param_items = pinfo.get("items")
            if param_items is not None:
                param_type = param_items["type"]
                params["properties"][pname]["items"] = {
                    "type": type_map[get_origin(param_type) or param_type],
                }

            if pinfo["required"]:
                required_params.append(pname)

        if required_params:
            params["required"] = required_params

        return params

    def get_auth(
        self,
        region,
        aws_access_key_id,
        aws_secret_access_key,
        aws_session_token,
        aws_bearer_token_bedrock,
    ) -> AwsAccessKeyAuth | AwsBearerTokenAuth | AwsFailedAuth:
        region = (
            region
            or os.getenv("AWS_REGION")
            or os.getenv("AWS_DEFAULT_REGION")
            or "us-west-2"
        )

        bearer_token = aws_bearer_token_bedrock or os.getenv("AWS_BEARER_TOKEN_BEDROCK")
        if bearer_token:
            return AwsBearerTokenAuth(
                aws_bearer_token_bedrock=bearer_token,
                region_name=region,
            )
        access_key_id = aws_access_key_id or os.getenv("AWS_ACCESS_KEY_ID")
        secret_access_key = aws_secret_access_key or os.getenv("AWS_SECRET_ACCESS_KEY")
        if access_key_id and secret_access_key:
            return AwsAccessKeyAuth(
                aws_access_key_id=access_key_id,
                aws_secret_access_key=secret_access_key,
                aws_session_token=aws_session_token or os.getenv("AWS_SESSION_TOKEN"),
                region_name=region,
            )
        return AwsFailedAuth(
            message="One of the following must be provided, neither was found:\n\n1. AWS_BEARER_TOKEN_BEDROCK\n2. AWS_ACCESS_KEY_ID + AWS_SECRET_ACCESS_KEY"
        )
